module A4Tools
  class BuiltinCommand
    attr_accessor :input, :error, :output, :args, :shell, :params
    attr_reader :status, :result

    class << self
      attr_reader :category_text, :description_text, :usage_text, :help_text, :validator, :tab_completors, :run_block, :opts, :is_user_defined
      attr_accessor :sha1, :filepath, :command

      def setup
        @description_text = ""
        @usage_text = ""
        @help_text = ""
        @category_text = "General"
        @validator = proc { true }
        @tab_completors = { default: proc { [] } }
        @run_block = proc { "" }
        @opts = []
      end

      ## Command category.
      ## Used to sort items in help.
      def category(category)
        @category_text = category
      end

      ## Summary of command
      ## One-line description of command for use in help
      def description(description)
        @description_text = description
      end

      ## Command line usage
      ## Added to end of usage argment. e.g. return "arg1 arg2" to get a string like "usage: my_command arg1 arg2"
      def usage(usage)
        @usage_text = usage
      end

      ## Help text
      ## Supplemental help added to end of --help.
      def help(help)
        @help_text = help
      end

      ## Command line validation
      ## Return true to indicate command line is valid.
      ## Return false to generate error and return usage help.
      ## Return a string to generate error and display the returned string.
      def validate(&validator)
        @validator = validator
      end

      ## Tab completion
      ## Return an array of possible values. You are not responsible for filtering these values for those matching the current argument.
      ## If an index is supplied, the block will only be invoked for that argument.
      def tab(index=nil, &tab_completor)
        if index.nil? then
          @tab_completors[:default] = tab_completor 
        else
          @tab_completors[index] = tab_completor
        end
      end

      ## Defines a command line argument. Identical syntax to Trollop.
      ##  opt :name, "Supplies a name", :type => :string
      def opt(key, description, params={})
        @opts.push({ key:key, description:description, params:params })
      end

      ## Actual code block to execute when the command is run
      def run(&block)
        @run_block = block
      end

      ###
      # Load from a file (don't call this directly)
      def load_command(file, userdefined=false)
        cmdname = File.basename(file.to_s, ".rb")
        subclass = Class.new(A4Tools::BuiltinCommand) {}
        contents = IO.read(file)

        subclass.setup
        subclass.command = cmdname
        subclass.filepath = file
        subclass.sha1 = Digest::SHA1.hexdigest contents
        subclass.user_defined if userdefined
        subclass.class_eval( contents, file )
        subclass
      end

      # Identify a command as user-defined (called automatically)
      def user_defined
        @is_user_defined = true
      end
    end


    ####
    # convenience methods you might want to be aware of

    def usher
      if @shell.shared[:usher].nil? then
        @shell.shared[:usher] = A4Tools::UsherClient.new()
      end
      @shell.shared[:usher]
    end

    def usherm
      @shell.client(:usherm)
    end

    def client
      @shell.active_client
    end

    def deployment
      @shell.client(:deployment)
    end

    def qa
      @shell.client(:qa)
    end

    def kaiconfig
      @shell.client(:kaiconfig)
    end

    def show_error(msg)
      @error.write "!!".style(:error) + " #{args[0].style(:command)}: #{msg}\n"
      return "", 1
    end

    def show_exception(exception)
      @error.write "\n"
      @error.write "!!".style(:error) + "#{command} crashed: #{exception.class.to_s} #{exception.message}\n"
      @error.write "Stack trace: #{exception.backtrace.map {|l| "  #{l}\n"}.join}"
      ""
    end

    ####
    # don't override these unless you really know what you're doing and have a good reason

    def initialize(shell, input=nil, output=nil, error=nil)
      @shell = shell
      @parser = parser

      @input = input || StandardInput.new
      @output = output || StandardOutput.new
      @error = error || StandardOutput.new
    end

    def category
      self.class.category_text
    end

    def command
      self.class.command
    end

    def description
      self.class.description_text
    end

    def usage
      self.class.usage_text
    end

    def help
      self.class.help_text
    end

    def validate
      block_call(&self.class.validator)
    end

    def user_defined?
      self.class.is_user_defined
    end

    def block_call(&block)
      self.class.send(:define_method, :__block_call, &block)
      self.class.send(:instance_method, :__block_call).bind(self).call
    end

    def sym_for_param(param)
      param[2..-1].to_sym rescue ""
    end

    def tab
      # only include --command-arguments if we already have a dash
      param_mix = []
      param_mix = all_opts.map { |opt| "--" + opt[:key].to_s.gsub("_","-") } if @realargs.last[0] == '-'

      key = if @realargs.length >= 2 and self.class.tab_completors.has_key? sym_for_param(@realargs[-2])
        sym_for_param(@realargs[-2])
      elsif self.class.tab_completors.has_key?(args.length-2)
        args.length - 2
      else
        :default
      end

      param_mix + block_call(&(self.class.tab_completors[key]))
    end

    def perform
      return [educate, 1] if params[:help]

      errmsg = validate
      return show_error(errmsg) if errmsg.is_a? String
      return show_error(usage_msg) if errmsg.nil? or errmsg == false

      begin
        block_call(&self.class.run_block)
      rescue Interrupt
        @error.write "Killed\n"
        return ["", 1]
      rescue SystemExit # allow commands to call exit
        raise
      rescue Exception => exception
        return show_exception(exception)
      end
    end

    def execute(input, output, error, args)
      @input = input
      @output = output
      @error = error
      parse(args)

      r = perform
      r = [r, 0] unless r.is_a? Array
      @result = r[0] || ""
      @status = r[1]

      txt = @result.to_s
      unless txt.empty? then
        txt += "\n" unless txt.end_with? "\n"
        @output.write txt
      end

      @status
    end

    def global_opts
      [ { key: :help, description:"Show help", params:{} } ]
    end

    def all_opts
      global_opts + self.class.opts
    end

    def parser
      Trollop::Parser.new(all_opts) do |opts|
        opts.each do |option|
          opt option[:key], option[:description], option[:params]
        end
      end
    end

    def parse(args)
      @realargs = Array.new(args)

      begin
        @params = @parser.parse(args)
      rescue Trollop::HelpNeeded
        @params = {help:true}
      rescue Trollop::VersionNeeded
        @params = {help:true}
      rescue Trollop::CommandlineError
        @params = {help:true}
      end
      @args = @parser.leftovers
      @args = args if @args.nil? or @args.empty?
    end

    def usage_msg
      u = usage
      if usage.is_a? Array then
        "usage:\n" + (u.map { |line| "\t#{command} [flags] #{line}" }).join("\n")
      else
        "usage: #{command} [flags] #{usage}"
      end
    end

    def educate
      buffer = PipeBuffer.new
      @parser.educate buffer
      intro_line = "#{command.style(:command)}: #{description}"
      userdef_line = user_defined? ? "User defined".style(:userdefined) : nil

      top_section = [intro_line, category.style(:category), userdef_line ].compact.join("\n")

      [top_section+"\n", usage_msg+"\n", buffer.buffer, help].compact.join("\n")
    end

    def print(s="")
      @output.print s
    end

    def puts(s="")
      @output.puts s
    end

    def console_puts(s)
      $stdout.puts s.gsub("\n","\r\n")
    end

    def console_print(s)
      $stdout.print s.gsub("\n","\r\n")
    end
  end
end
