category "Protocol manipulation"
description "Interactively build a JSON object of a given class"
usage "class_name"
help <<-EOS
Launches an interactive prompt to build the indicated Talk class. At the conclusion of this prompt, a JSON string will be printed to stdout.

This command is somewhat magical in that the interactive prompts do not appear on the command's standard output, nor does it gather prompt input from standard input; only the JSON appears on standard output.

Therefore, you can pipe the output of #{command.style(:command)} directly to #{"send".style(:command)} or some other command.
EOS


validate { args.length == 2 }
tab 0 do
  talk.things_named_like(args[1], :class).map { |thing| talk.truncated_name thing[:name] }
end

run do
  cls = talk.class_named(args[1])
  return show_error("#{args[1]}: no such class defined") if cls.nil?

  obj = build(cls)
  obj[:__class] = talk.class_named(args[1])[:name] rescue args[1]
  JSON.generate(obj)
end

###

def prompt_for_path(path)
  path.join(".").style(:class) + ") "
end

def prompt_for_value(path, type)
  "#{path.join(".").style(:class)} #{type.style(:field_datatype)}) "
end

def get_action(allowed)
  char = STDIN.gets.chomp rescue nil
  char = nil unless char.nil? or allowed.include? char
  char
end

def get_value_real(value)
  case value
  when nil, "null", ""
    nil
  else
    value.to_f
  end
end

def get_value_integer(value)
  case value
  when nil, "null", ""
    nil
  else
    value.to_i
  end
end

def get_value_boolean(value)
  case value
  when "0", "false"
    false
  when "1", "true"
    true
  when "", "null"
    nil
  else
    true
  end
end

def get_value_string(value)
  case value
  when nil, "null"
    nil
  else
    value
  end
end

def get_value(field, path)
  type = field[:type].first
  console_print prompt_for_value(path, type)

  value = STDIN.gets
  if value.nil? then
    value = "null" # nil => CTRL+D. Take advantage that all get_value_* interpret null as nil-value to make CTRL+D -> null shortcut.
    console_puts # need a newline since none was actually typed
  end

  value = value.chomp
  if(value == "?") then
    show_field_help(field)
    return get_value(field, path)
  end

  return get_value_integer(value) if([ "uint8", "uint16", "uint32", "uint64", "int8", "int16", "int32", "int64" ].include?(type))
  return get_value_real(value) if([ "real" ].include?(type))
  return get_value_boolean(value) if([ "bool" ].include?(type))
  return get_value_string(value) if([ "string" ].include?(type))
end

def in_array?(field, depth)
  field[:type][-1 - depth] == "[]"
end

def in_hash?(field, depth)
  field[:type][-1 - depth] == "{}"
end

def in_container?(field, depth)
  in_array?(field, depth) or in_hash?(field, depth)
end

def prompt_for_action(field, path, type, actions)
  while true do
    console_puts "#{path.join('.').style(:class)}: #{type.style(:field_datatype)}. #{actions.map { |a| "[#{a[0].bold}]#{a[1..-1]}" }.join(', ')}"
    console_print prompt_for_path(path)
    action = get_action(actions.map { |a| a[0] })
    return (actions.select { |a| a[0] == action }).first unless action.nil?
  end
end

def type_at_depth(field, depth)
  field[:type][0 .. -1 - depth].join("")
end

def show_field_help(field, depth=-1)
  depth = field[:type].length - 1 if depth < 0
  console_puts "#{talk.truncated_name(type_at_depth(field, depth)).style(:field_datatype)} #{field[:name].style(:field_name)}\n\t#{field[:description]}"
end

def handle_array_start(field, path, depth)
  actions = [ "build", "null", "?" ]

  case prompt_for_action(field, path, type_at_depth(field, depth), actions)
  when "?"
    show_field_help(field, depth)
    return handle_array_start(field, path, depth)
  when nil, "null"
    return nil
  when "build"
    collection = []
    building = true
    index = 0
    while true do
      element = handle_array_element(field, path[0..-2] + ["#{path.last}[#{index}]"], depth+1)
      return collection if element.nil?
      collection.push element[:value]
      index += 1
    end
  end
end

def handle_array_element(field, path, depth)
  actions = [ "build", "null", "done", "?" ]
  case prompt_for_action(field, path, type_at_depth(field, depth), actions)
  when "?"
    show_field_help(field, depth)
    return handle_array_element(field, path, depth)
  when nil, "null"
    return { value:nil }
  when "done"
    return nil
  when "build"
    return { value:handle_field(field, path, depth+1, true) }
  end
end

def populate(field, path)
  while true do
    idx = 0
    builders = ObjectBuilder.builders_for_name(field[:type].first)
    builders.each do |builder|
      console_puts("#{idx.to_s.style(:constant_value)} #{builder.identifier.to_s.style(:method_name)}: #{builder.description}")
      idx += 1
    end

    console_print prompt_for_path(path)
    action = STDIN.gets.chomp rescue nil

    return handle_complex(field, path) if action.nil? # CTRL+D cancels populate
    index = action.match(/^[0-9]+$/) ? action.to_i : nil
    return builders[index].value if not index.nil? and index < builders.length
  end
end

def has_cache?(field)
  ObjectBuilder.supported_classes(false).each do |name|
    return true if talk.name_matches?(name, field[:type].first)
  end
  false
end

def handle_complex(field, path, assume_build=false)
  return build(talk.class_named(field[:type].first), path) if assume_build

  actions = if has_cache?(field) then
    [ "build", "populate", "null", "?" ]
  else
    [ "build", "null", "?" ]
  end

  case prompt_for_action(field, path, field[:type].first, actions)
  when "build"
    build(talk.class_named(field[:type].first), path)
  when nil, "null"
    nil
  when "populate"
    populate(field, path)
  when "?"
    show_field_help(field, field[:type].length-1)
    return handle_complex(field, path)
  end
end

def handle_primitive(field, path)
  get_value(field, path)
end

def handle_field(field, path, depth=0, assume_build=false)
  path += [field[:name]] if depth == 0
  if depth+1 < field[:type].length then
    if in_array?(field, depth)
      return handle_array_start(field, path, depth)
    else
      # TODO: handle dictionary
      return nil
    end
  end

  if talk.primitive? field[:type].first
    return handle_primitive(field, path)
  else
    return handle_complex(field, path, assume_build)
  end
end

def build(cls, path=[])
  if path.empty? then
    console_puts "Building #{talk.truncated_name cls}..."
  else
    console_puts "#{talk.truncated_name cls} #{path.last}\n\t#{cls[:description]}"
  end

  result = {}
  cls[:field].each { |field| result[field[:name]] = handle_field(field, path) }
  result
end
