category "Protocol manipulation"
description "Connect to a remote server"
usage "url [client_name]"

help <<-EOS
Connect to a remote server, indicated by the URL.
Supported schemes: http, https, ws, wss.

Once connected, commands like #{"send".style(:command)} will use the connected server as a target. Connecting also resets the #{"recap".style(:command)} log.

Comet support is not included; therefore, http and https provide synchronous communication only. ws and wss support fully asynchronous communication.

#{command.style(:command)} does make special effort to validate that a URL is accessible. For http and https clients, this means that attempts to use this connection will fail. Due to the nature of websockets, #{command.style(:command)} will display an error on exit if these services are unavailable.
EOS

validate { args.length >= 2 }

opt :kai, "Connect the Kai Mercury, Central and QATest clients to a particular site ID or host"

tab 0 do
  if params[:kai] then
    (usherm[:usherSystemId][:usherSites].map { |site| site[:idSite] } +
      usherm[:usherSystemId][:usherSites].map { |site| site[:name] }) rescue []
  else
    []
  end
end

run do
  if params[:kai] then
    connect_kai
  else
    connect_raw
  end
end

###

def connect_raw
  key = args[2].to_sym rescue :user
  clients = []
  client = AcresClient.new(args[1])
  return show_error("Unable to contact server at #{args[1]}") unless client.transport_connect
  
  @shell.set_client(key, client)
  client.on(:connect) { |trigger, msg| @shell.notify_connect(client) }
  client.on(:sent) { |trigger,msg| @shell.notify_outgoing_message(client, msg) }
  client.on(:message) { |trigger,msg| @shell.notify_incoming_message(client, msg) }  
  client.on(:disconnect) { |trigger,msg| @shell.notify_disconnect(client) }

  @shell.set_active_client(key)
  ""
end

def connect_kai
  dest = args[1]
  
  if dest.match(/^([a-zA-Z]+):\/\/[a-zA-Z0-9\-\.]+(:[0-9]+)?(\/)?/) then
    connect_to_url(dest)
  elsif dest.match(/^[0-9]+$/) then
    connect_to_site_id(dest.to_i)
  else
    connect_to_usher_host(dest)
  end
end

def request_caches
  version = @shell.client(:config).site_version
  client_interfaces.keys.each { |key| @shell.client(key).version = version }
  @shell.client(:config).machines
end

def connect_to_url(base)
  client_interfaces.each do |key, uri|
    url = URI.join(base, uri).to_s
    @shell.set_client(key, client_for_key(key, url))
  end

  request_caches
  "Connected to #{@shell.client(:config).version.style(:host)} server"
end

def connect_to_site_id(site_id)
  site = usherm.site_with_id(site_id)
  return show_error("Site ##{site_id}: No such site ID in usher") if site.nil?
  provider = usherm.provider_for_product(site_id, "Mercury")
  return show_error("Site ##{site_id}: Does not offer Mercury") if provider.nil?
  url = usherm.url_for_provider(provider)
  return show_error("Site ##{site_id}: Site appears to be down") if url.nil?
  puts "Connecting to Kai at #{url}"
  connect_to_url(url)
end

def connect_to_usher_host(host)
  site = usherm.site_with_name(host)
  return show_error("Site #{host}: Does not appear to be listed in usher") if site.nil?
  connect_to_site_id(site[:idSite])
end

def client_for_key(key, url)
  case key
  when :mercury
    AcresClient.new(url, @shell.get_env(:kai_user), @shell.get_env(:kai_pass))
  when :central
    AcresClient.new(url, @shell.get_env(:usherm_user), @shell.get_env(:usherm_pass))
  when :config
    KaiConfigClient.new(url, @shell.get_env(:usherm_user), @shell.get_env(:usherm_pass))
  when :qa
    AcresClient.new(url, @shell.get_env(:kai_user), @shell.get_env(:kai_pass))
  else
    AcresClient.new(url, @shell.get_env(:usherm_user), @shell.get_env(:usherm_pass))
  end
end

def client_interfaces
  {
    mercury:"/mercury/client?wrap",
    central:"/central/central?wrap",
    config:"/config/config?wrap",
    qa:"/mercury/qatest?wrap"
  }
end
