require 'date'

category "Deployment"
description "Manage deployments"
usage [
  "list          -- list active publications",
  "builds        -- list builds",
  "publish       -- publish a build"
]

opt :url, "Deployment server URL"

opt :app, "List builds for application", :type => :string
opt :jeb, "List builds for jeb (i.e. three-word name)", :type => :string
opt :branch, "List builds for branch", :type => :string
opt :commit, "List builds for commit hash", :type => :string
opt :yes, "Automatically include all apps in publication. No effect if --app is supplied."

opt :site, "When used with list, only list publications to given site. When used with publish, indicates site to deploy to.", :type => :integer


validate do
  return false if args.length <= 1

  case args[1]
  when "builds"
    args.length == 2
  when "list"
    true
  when "publish"
    return "Must supply at least one site ID" unless args.length >= 3
    true
  else
    "#{args[1]}: unknown subcommand"
  end
end

tab(:app) { all_apps }
tab(:jeb) { all_jebs }
tab(:branch) { all_branches }
tab(:commit) { all_commits }

tab 0 do
  [:list, :builds, :publish]
end

run do
  usherm.ensure_fresh
  deployment.ensure_fresh
  case args[1]
  when "builds"
    builds
  when "list"
    list
  when "publish"
    publish
  end
end

###
# Tab completors

def all_apps
  all_of_key(:appId)
end

def all_branches
  all_of_key(:branches)
end

def all_commits
  all_of_key(:commitId)
end

def all_jebs
  all_of_key(:buildId)
end

def all_of_key(key)
  instances = {}

  deployment[:buildList][:builds].each do |build|
    if build[key].is_a? Array then
      build[key].each { |item| instances[item] = true }
    else
      instances[build[key]] = true
    end
  end

  instances.keys
end

###
# List

def list
  sites = if args.length >= 3 then
    args[2..-1].map { |arg| arg.to_i }
  else
    (usherm[:usherSystemId][:usherSites].map { |site| site[:idSite] }).sort
  end

  sites.each do |site_id|
    list_site(site_id, deployment.query(:buildPublicationList, site_id))
  end
  ""
end

def list_site(site_id, publications)
  site = usherm.site_with_id(site_id) || {name:"Unknown site", siteDescription:"Not in usher"}
  lines = []
  
  lines.push "Site #{site_id} -- #{site[:name]} (#{site[:siteDescription]})"
  publications[:orders].keys.sort.each do |app|
    next if params[:app] and not talk.name_matches?(app, params[:app])
    lines.push list_site_app(app.to_s, publications[:orders][app])
  end

  puts lines.join("\n")+"\n" unless lines.length == 1
end

def list_site_app(app, orders)
  pub = orders.first
  return if params[:jeb] and pub[:buildInfo][:buildId] != params[:jeb]
  return if params[:commit] and pub[:buildInfo][:commitId] != params[:commit]
  return if params[:branch] and not(pub[:buildInfo][:branches].include? params[:branch])
  "\t#{timestamp(pub[:publicationTime])} #{app.style(:class)} -- #{build_info(pub[:buildInfo])}"
end

###
# Builds

def builds
  show_build_set(build_set)
  ""
end

def build_set
  if params[:jeb] then
    builds_for_jeb(params[:jeb])
  elsif params[:commit] then
    builds_for_commit(params[:commit])
  elsif params[:branch] then
    builds_for_branch(params[:branch])
  elsif params[:app] then
    builds_for_app(params[:app])
  else
    deployment[:buildList][:builds]
  end
end

def builds_for_app(app)
  deployment[:buildList][:builds].select { |build| talk.name_matches?(build[:appId], app) }
end

def builds_for_commit(commit)
  set = 
  if params[:app] then
    builds_for_app(params[:app])
  else
    deployment[:buildList][:builds]
  end

  set.select { |build| build[:commitId].start_with? commit }
end

def builds_for_jeb(jeb)
  set = 
  if params[:app] then
    builds_for_app(params[:app])
  else
    deployment[:buildList][:builds]
  end
  
  set.select { |build| build[:buildId] == jeb }
end

def builds_for_branch(branch)
  set = 
  if params[:app] then
    builds_for_app(params[:app])
  else
    deployment[:buildList][:builds]
  end
  
  set.select { |build| build[:branches].include? branch }
end

def build_set_by_app(builds)
  apps = {}
  builds.each do |build|
    apps[build[:appId]] ||= []
    apps[build[:appId]].push build
  end
  apps
end

###
# Publish

def prompt_for_apps(set)
  apps = build_set_by_app(set)
  return apps.keys.sort.map { |app| apps[app].last } if params[:yes]

  selected = []
  apps.keys.sort.each do |app|
    build = apps[app].last
    cls = sprintf("%s", app).style(:class)
    name = sprintf("%s", build[:buildId]).style(:build_name)
    branch = build[:branches].join(",").style(:branch)
    line = sprintf("%99s", "#{cls}: #{name} [#{branch}]")

    while true do
      print "#{line} [y/N]> "
      input = STDIN.gets.chomp.upcase rescue nil
      input = "N" if not input.nil? and input.empty?

      selected.push build if input == "Y"
      puts if input.nil?
      break if ["Y", "N"].include? input
    end
  end

  selected
end

def publish
  sites = args[2..-1].map { |arg| arg.to_i }
  selected = prompt_for_apps(build_set)
  selected.each do |build|
    puts "Publishing #{build[:appId].style(:class)} \"#{build[:buildId].style(:build_name)}\" to site #{sites.join(",")}..."
    deployment.publish_build(build, sites)
  end
  "Done."
end

###
# Other stuff
def show_app(app, limit=50)
  list = deployment.index_list_by_app[app]
  return show_error("#{app}: No such app in deployment server") if list.nil?

  list.each do |build|
    puts show_build(build)
  end

  ""
end

def timestamp(ts)
  DateTime.strptime(ts.to_s,'%s').strftime("%Y-%m-%d %H:%M:%S").style(:timestamp)
end

def build_id(build)
  sprintf("%4d", build[:buildRecordId]).style(:identifier)
end 

def branches(build)
  build[:branches].join(",").style(:branch)
end

def author(build)
  build[:author].style(:author) rescue ""
end

def build_info(build)
  name = sprintf("%36s", build[:buildId]).style(:build_name)
  "#{name} [#{branches(build)}] -- #{author(build)}"
end

def show_app(app, builds)
  puts "#{app.style(:class)}: #{builds.length} builds"
  builds.each { |build| puts timestamp(build[:postDate]) + " " + build_info(build) }
  puts
end

def show_build_set(set)
  apps = build_set_by_app(set)
  apps.keys.sort.each { |app| show_app(app, apps[app]) }
end
