category "Usher"
description "Show information about Usher hosts"
usage "hostname"
help <<-EOS
Shows information for a given usher host. If no usher host is specified, lists all hosts.
EOS

validate do
  return show_error("must connect usherm first") if usherm.nil?
  args.length < 3
end

tab 0 do
  (usherm[:usherSystemId][:systemHosts].map { |host| host[:hostName] }) rescue []
end

run do
  usherm.ensure_fresh
  if args.length == 1 then
    (usherm[:usherSystemId][:systemHosts].map do |host|
      concise_host(host)
    end).join("\n")
  elsif args.length == 2 then
    host = usherm.host_named(args[1])
    return show_error("#{args[1]}: unknown host") if host.nil?
    verbose_host(host)
  end
end

###

def string_for_range(start, last, params={})
  if start == last then
    last
  else
    "#{start}-#{last}"
  end
end

def collapse_ranges(elements)
  return "None" if elements.nil? or elements.empty?
  elements = (elements.map { |element| element.to_i }).sort

  ranges = []
  start = elements.first
  last = nil

  elements.each do |element|
    if not(last.nil?) and element != last + 1
      ranges.push(string_for_range(start, last))
      start = element
    end
    last = element
  end

  ranges.push(string_for_range(start, last))
  ranges.join(", ")
end

def provider_line(id_provider)
  provider = usherm.provider(id_provider)
  return "Non-existent provider #{id_provider}" if provider.nil?
  return nil if provider[:isSpecial]

  sites = collapse_ranges(provider[:idSites])

  ipLines = ["Internal", "External"].map do |zone|
    hosts = provider["usherHosts#{zone}".to_sym] || []
    zone + ": " + if hosts.empty? then
      "None"
    else
      urls = hosts.map { |host| host[:url] }
      urls.join(", ")
    end
  end

  out = "\t\t#{provider[:product].style(:product)}: Site #{sites}\n"
  out += (ipLines.map { |line| "\t\t\t#{line}"}).join("\n") + "\n"
  out
end

def acl_line(acl)
  service = usherm.service_for_id(acl[:idService])
  return "Non-existent service #{acl[:idService]}" if service.nil?

  proto = service[:isUdp] ? "UDP" : "TCP"
  external_addr = [acl[:ip], acl[:port]].compact.join(":")
  "\t\t#{service[:name].style(:service)}, #{proto} #{service[:port]} (#{external_addr} external)"
end

def verbose_host(host)
  datacenter = usherm.datacenter(host[:idDataCenter])
  out = "Host #{host[:hostName].style(:host)} (#{host[:idSystemHost]})\n"
  out += "\tInternal #{host[:ipAddrInt].style(:ip)}\n"
  out += "\tExternal #{host[:ipAddrExt].style(:ip)}\n"
  if datacenter.nil?
    out += "\tDatacenter: not found in usher (#{host[:idDataCenter]})\n"
  else
    out += "\tDatacenter: #{datacenter[:dataCenterName].style(:datacenter)} (#{host[:idDataCenter]})\n"
  end
  out += "\tHost environment: #{usherm.env_type(host[:systemHostEnvironment])}\n"
  out += "\tUsher system type: #{usherm.system_type(host[:usherResponseType])}\n"
  out += (host[:providers].map { |id_provider| provider_line(id_provider) }).compact.join("\n")

  out += "\tACLs:\n"
  acls = usherm.acls_for_host_id(host[:idSystemHost])
  if acls.empty?
    out += "\t\tNo ACLs for this host."
  else
    out += (acls.map { |acl| acl_line(acl) }).compact.join("\n")
  end

  out
end

def concise_host(host)
  datacenter = usherm.datacenter(host[:idDataCenter])
  dcnote = if datacenter.nil?
    "unknown (#{host[:idDataCenter]})"
  else
    "#{datacenter[:dataCenterName].style(:datacenter)}"
  end

  "#{host[:hostName].style(:host)} -- #{dcnote} #{usherm.env_type(host[:systemHostEnvironment])}"
end

