category "Protocol manipulation"
description "Access a pre-cached request object"
usage "[identifier]"
help <<-EOS
Reading cache
#{command.style(:command)} builds object instances based on definitions. These definitions can either be Ruby scripts providing dynamic content, or JSON strings with static content.

The definitions included by default are located in lib/object_builder/definitions, and the filenames reflect the identifiers used by #{command.style(:command)}. Adding a file to this directory will cause it to be included automatically. Files ending in .rb will be interpreted as ruby scripts; all others will be treated as raw JSON.

Writing cache
Using the --store option #{command.style(:command)} can write objects to the cache.
EOS

validate do
  return args.length <= 3 if params[:store]
  return args.length == 2 if params[:persist]
  args.length <= 2
end

tab 0 do
  (ObjectBuilder.identifiers_like(args[1]).map { |id| id.to_s }) + ObjectBuilder.supported_classes
end

opt :store, "Store a JSON object supplied on STDIN to the specified identifier", :type => :string
opt :persist, "Write an identifier to disk"
opt :field, "Read a specific field from the JSON string", :type => :string
opt :type, "Assume the top-level datatype is of a specified type", :type => :string

run do
  return store if params[:store]
  return persist if params[:persist]
  read
end

###

def acquire_object
  obj = JSON.parse(@input.read_all) rescue nil
  talk.locate_data(obj, params[:field], params[:type]) rescue nil
end

def store
  description = args[1] rescue "Defined from console"
  obj, type = acquire_object
  return show_error("Unable to acquire field from JSON") if obj.nil?

  obj[:__class] = type unless type.nil?

  ObjectBuilder.load_object(obj, params[:store], description)
  ""
end

def persist
  builder = ObjectBuilder[args[1]]
  return show_error("#{args[1]}: no such identifier") if  builder.nil?
  File.write(File.join(@shell.shelldir(:json), args[1]), builder.value.to_json)
  ""
end

def read
  if args.length == 1 then
    builders = ObjectBuilder.builders
    return show_error("no builders defined") if builders.empty?
    return show_builder_list(builders)
  end

  builder = ObjectBuilder[args[1]]
  return builder.value.to_json unless builder.nil?

  builders = ObjectBuilder.builders_for_name(args[1])
  return "#{args[0]}: #{args[1]}: no such identifier or class in cache" if builders.empty?
  show_builder_list(builders)
end

def show_builder_list(builders)
  (builders.map do |builder|
    "#{builder.name.to_s.style(:class)} #{builder.identifier.to_s.green}: #{builder.description}"
  end).join("\n")
end

