category "Protocol manipulation"
description "Print message history for current connection"
usage [
  "#{command}             show all messages",
  "#{command} n           show nth message before most recent",
  "#{command} start end   show messages from start to end"
]

help <<-EOS
Examples:
List all messages
 #{command}

List most recent message
 #{command} 0

List most recent 3 messages
 #{command} 2 0

List first message in conversation
 #{command} -1

List first 5 messages in conversation
 #{command} -5 -1
EOS

opt :client, "Recap this history for a client other than the active client", :type => :string
opt :summary, "Only list class names in text mode"
opt :json, "Output as JSON"
opt :raw, "Don't wrap JSON objects with metadata objects."
opt :array, "Force JSON to always be an array, even if only a single object is returned"
opt :reverse, "Index messages from end of conversation (so message 0 is most recent)"

validate do
  return "Must have an active client" if client.nil? and not params[:client]
  return "#{params[:client]}: unknown client" if params[:client] and not @shell.clients.include? params[:client].to_sym
  true
end

tab :client do
  @shell.clients
end

run { use_json? ? json_output : text_output }

###

def target_client
  params[:client] ? @shell.client(params[:client]) : client
end

def messages
  messages = target_client.history
end

def range_for_args(args)
  case args.length
  when 0,1
    range = (0 .. -1)
  when 2
    range = args[1].to_i .. args[1].to_i
  when 3
    low = [args[1], args[2]].min
    high = [args[1], args[2]].max
    range = low.to_i .. high.to_i
  end
end

def sender(msg)
  case msg[:sender]
  when :client
    "client".light_blue
  when :server
    "server".magenta
  end
end

def timecode(msg)
  delta = msg[:time] - target_client.start_time
  all_seconds = delta.to_i

  milliseconds = 1000.0*(delta - all_seconds)
  hours = (all_seconds / 3660).to_i
  minutes = ((all_seconds - 3660*hours) / 60).to_i
  seconds = (all_seconds - 3600*hours - 60*minutes).to_i

  return sprintf("%3d:%02d:%02d.%03d", hours, minutes, seconds, milliseconds).light_black
end

def text_output
  index = -1
  set = params[:reverse] ? messages : messages.reverse
  (set.reverse[range_for_args(args)].map do |msg|
    index += 1
    index_str = sprintf("%3d", index)
    "#{index_str} #{timecode(msg)} #{sender(msg)} #{render_text(msg[:message])}"
  end).join("\n")
end

def json_output
  set = params[:reverse] ? messages : messages.reverse
  set = (set.reverse[range_for_args(args)].map do |msg|
    if params[:raw] then
      msg[:message]
    else
      msg.reject { |k,v| k == :raw }
    end
  end)

  set = set[0] if set.length == 1 and not params[:array]
  set.to_json
end

def use_json?
  params[:json] or params[:raw] or params[:array]
end

def render_text(message)
  if params[:summary] then
    render_summary(message)
  else
    message.is_a?(Hash) ? JSON.generate(message) : message
  end
end

def render_summary(message)
  toplevel, cls = pick_toplevel(message)
  puts toplevel if cls.nil?
  cls ||= talk.guess_class(toplevel)
  class_name = cls.style(:class) || "Unknown"
  id_str = message[:id] ? sprintf("%3d", message[:id]) : "   "

  sprintf("%s %20s %s %s",
    id_str.style(:id),
    method(message).style(:method_name),
    class_name.style(:class),
    human_readable_size(message.to_json.length)
    )
end

def human_readable_size(size)
  titles = [ "B", "KiB", "MiB", "GiB", "TiB" ]
  titles.each_index do |idx|
    title = titles[idx]
    unit = 1024**idx
    boundary = 1024*unit
    return "#{size/unit} #{title}" if size < boundary or title == titles.last
  end
end

def method(message)
  return message[:method] if message.has_key? :method
  request = find_request_with_id(message[:id])
  request[:method] rescue "unknown"
end

def find_request_with_id(id)
  target_client.history.each do |msg|
    next if msg[:message].nil? or msg[:message][:id].nil? or not(msg[:message].has_key? :params)
    return msg[:message] if msg[:message][:id].to_i == id.to_i
  end

  nil
end

def pick_toplevel(message)
  message = symbolify(JSON.parse(message)) if message.is_a? String
  if message[:jsonrpc] == "2.0" then
    return bypass(message, "com.acres4.common.info.JSONRPCRequest") if message.has_key? :params
    return [message, "Error"] unless message[:error].nil?
    return bypass(message, "com.acres4.common.info.JSONRPCResponse") if message.has_key? :result
  end

  [message, talk.guess_class(message)]
end

def bypass(object, type)
  case type
  when "com.acres4.common.info.NamedObjectWrapper"
    return bypass(object[:body], object[:className])
  when "com.acres4.common.info.JSONRPCRequest"
    return bypass(object[:params], talk.guess_class(object[:params]))
  when "com.acres4.common.info.JSONRPCResponse"
    return bypass(object[:result], talk.guess_class(object[:result]))
  else
    return [object, type]
  end
end
