category "Protocol manipulation"
description "Send a request to the remote server"

usage [
  "                  send a raw message to the server",
  "interface method  send a message to the server, wrapped with a JSON-RPC request for the specified method"
]

validate { args.length == 1 || args.length == 3 }

help <<-EOS
Messages are supplied as JSON via stdin. Messages can therefore be read using input redirection or piped from another command, like #{"build".style(:command)} or #{"cached".style(:command)}.
EOS

tab 1 do
  set = talk.definition[:protocol].inject([]) do |set, proto|
    set + proto[:method].select { |method| talk.name_similar?(method[:name], args[2] || "") }
  end

  set.map { |m| m[:name] }
end

opt :no_inject, "Don't automatically inject the current connection token"
opt :dry, "Don't send the message; just print what would be sent to the console"
opt :spam, "Send the message repeatedly", :type => :integer, :default => 1
opt :spam_interval, "Time to wait between spams in milliseconds", :type => :integer, :default => 1000

run do
  return show_error("must connect first") if @shell.shared[:client].nil?
  @shell.allow_alerts

  message = input.read_all
  index = 0

  params[:spam].times do
    index = index + 1
    if args.length == 3 then
      send_wrapped(message)
    else
      send_raw(message)
    end

    puts "Sending message #{index} of #{params[:spam]}" if params[:spam] > 1
    sleep params[:spam_interval] * 1.0/1000.0 if index < params[:spam]
  end

  ""
end

def send_raw(message)
  send(message)
end

def send_wrapped(message)
  message = symbolify(JSON.parse(message)) if message.is_a? String
  method, protocol = talk.method_named(args[2])
  return show_error("#{args[1]}: unknown method") if method.nil?
  cls = talk.class_named(method[:request])
  return show_error("#{args[1]}: unable to locate request class") if cls.nil?

  client.inject_token(message) unless params[:no_inject]
  jsonrpc = client.jsonrpc_message(method[:name], args[1], talk.guess_class(message), message)

  message = JSON.generate(jsonrpc)
  send(message)
end

def send(message)
  if params[:dry] then
    puts message
  else
    client.send_message(message)
  end
end
