category "Talk"
description "Show details for a Talk entity"
usage [
  "entity                show information about an entity",
  "enumeration constant  show information about a constant in a particular enumeration",
  "glossary term         show information about a term in a particular glossary",
  "class field           show information about a class in a particular field",
  "protocol method       show information about a method in a particular protocol"
]

help "Displays Talk information about a specified entity. Entities can be classes, glossaries, enumerations, protocols, constants, fields, terms, or methods."

validate { args.length <= 3 }

tab 0 do
  talk.things_named_like(args.last).map { |thing| talk.truncated_name thing[:name] } rescue []
end

opt :class, "List all classes"
opt :enumeration, "List all enumerations"
opt :glossary, "List all glossaries"
opt :protocol, "List all protocols"
opt :recurse, "List all entities referenced by queried entity"
opt :reverse, "List entities in reverse (with original query at top)"

run do
  if args.length > 1 then
    @subqueries = []
    top_level = show(args[1 .. -1]) 
    return show_error("#{args[1 .. -1].join(' ')}: symbol not found") if top_level.nil?

    sections = [ top_level ]

    if params[:recurse] then
      until @subqueries.empty? do
        query = @subqueries.first
        @subqueries = @subqueries[1..-1]
        sections.push show([query])
      end
    end
    
    sections.reverse! unless params[:reverse]
    sections.each do |section|
      puts section
    end
    ""
  else
    show_list
  end
end

###

def add_queries(*new_queries)
  @subqueries ||= []
  @seen ||= {}
  new_queries.each do |q|
    next if q == "none" # given by methods
    @subqueries.push q unless @seen.has_key? q
    @seen[q] = true
  end
end

def show_list
  s = ""
  query_for_params.each do |section|
    next if talk.definition[section].nil?
    s += descriptions[section] + " (#{talk.definition[section].length})\n"
    s += (talk.definition[section].map { |item| "\t#{item[:name].style(style_for_section(section))}\n\t#{item[:description]}\n" }).join("\n") + "\n\n"
  end
  s
end

def query_for_params
  query = []
  options = [ :class, :enumeration, :glossary, :protocol ]
  options.each { |opt| query.push(opt) if params[opt] }

  query = options if query.empty?
  query
end

def style_for_section(section)
  case section
  when :class
    :class
  when :protocol
    :protocol_name
  when :enumeration, :glossary
    :collection_name
  else
    :none
  end
end

def descriptions
  { class:"Classes", glossary:"Glossaries", enumeration:"Enumerations", protocol:"Protocols" }
end

def list(query)
  query = [ :class, :enumeration, :glossary, :protocol ] if query.empty?
  query.map! { |e| e.to_sym }

end

##

def show_field(query)
  return nil if query.length < 2
  cls = (query[0].is_a? Hash) ? query[0] : talk.class_named(query[0])
  return nil if cls.nil?
  field = talk.field_named(cls, query[1])
  return nil if field.nil?

  add_queries(field[:type].first)

  "Field #{field[:name]}: #{(field[:type].map { |t| talk.truncated_name t }).join('')}\n\tClass #{cls[:name]}\n\t#{field[:description]}\n"
end

def show_class(query)
  cls = (query[0].is_a? Hash) ? query[0] : talk.class_named(query[0])
  return nil if cls.nil?

  fields = cls[:field].map do |f|
    add_queries(f[:type].first)
    "\t#{f[:name].style(:field_name)}, #{(f[:type].map { |t| talk.truncated_name t }).join('').style(:field_datatype)}\n\t\t#{f[:description]}\n"
  end

  "Class #{cls[:name].style(:class)}\n\t#{cls[:description]}\n\n" + fields.join("\n")
end

def origin_line(method)
  return "Asynchronous from server\n\t\t".style(:origin) if method[:origin] == "server"
  return ""
end

def show_protocol(query)
  proto = (query[0].is_a? Hash) ? query[0] : talk.protocol_named(query[0])
  return nil if proto.nil?

  methods = proto[:method].map do |m|
    add_queries(m[:request], m[:response])
    "\t#{m[:name].style(:method_name)}: #{m[:request].style(:class)} -> #{m[:response].style(:class)}\n\t\t#{origin_line(m)}#{m[:description]}\n"
  end

  "Protocol #{proto[:name].style(:protocol_name)}\n\t#{proto[:description]}\n" + methods.join("\n")
end

def show_method(query)
  method, protocol = (query.length == 1) ? talk.method_named(query[0], nil) : talk.method_named(query[1], query[0])
  return nil if method.nil?

  add_queries(method[:request], method[:response])
  "Method #{method[:name].style(:method_name)}: #{method[:request].style(:class)} -> #{method[:response].style(:class)}\n\tProtocol #{protocol[:name].style(:protocol_name)}\n\t#{method[:description]}\n\n"
end

def show_constant(query)
  constant, container = (query.length > 1) ? talk.constant_named(query[1], query[0]) : talk.constant_named(query[0], nil)
  return nil if constant.nil?

  origin = (container.has_key? :term) ? "Glossary #{container[:name]}" : "Enumeration #{container[:name]}"

  "Constant #{constant[:name].style(:constant_name)}: #{constant[:value].to_s.style(:constant_value)}\n\t#{origin}\n\t#{constant[:description]}\n\n"
end

def show_glossary(query)
  glossary = (query[0].is_a? Hash) ? query[0] : talk.glossary_named(query[0])
  return nil if glossary.nil?

  "Glossary #{glossary[:name].style(:collection_name)}\n\t#{glossary[:description]}\n\n" + (glossary[:term].map { |t| "\t#{t[:name].style(:constant_name)} (\"#{t[:value].style(:constant_value)}\")\n\t\t#{t[:description]}\n"}).join("\n")
end

def show_enumeration(query)
  enumeration = (query[0].is_a? Hash) ? query[0] : talk.enumeration_named(query[0])
  return nil if enumeration.nil?

  "Enumeration #{enumeration[:name].style(:collection_name)}\n\t#{enumeration[:description]}\n\n" + (enumeration[:constant].map { |c| "\t#{c[:value].to_i.to_s.style(:constant_value)} #{c[:name].style(:constant_name)}\n\t\t#{c[:description]}\n"}).join("\n")
end

def show(query)
  show_field(query) || show_method(query) || show_class(query) || show_protocol(query) || show_constant(query) || show_glossary(query) || show_enumeration(query)
end

