module A4Tools
  class ObjectBuilder
    class << self
      attr_reader :is_cached, :generate_block, :description_text
      attr_accessor :id

      def load_objects
        @definitions = {}
        load_path(File.join(File.dirname(__FILE__), "definitions"))
      end

      def load_path(path)
        if File.directory?(path) then
          define_with_directory(path).each do |defn|
            @definitions[defn.identifier] = defn
          end
          true
        elsif File.file?(path) then
          defn = define_with_filename(path)
          @definitions[defn.identifier] = defn
          true
        else
          false
        end
      end

      def load_object(obj, id, description)
        obj = JSON.generate(obj) unless obj.is_a? String
        @definitions[id.to_sym] = define_with_json(obj, id.to_sym, description)
      end

      def define_with_directory(directory)
        Dir[File.join(directory, "*")].map { |file| define_with_filename(file) }
      end

      def define_with_filename(filename)
        contents = IO.read(filename)
        id = File.basename(filename).gsub(/\..*$/, "").to_sym
        return define_with_script(contents, id) if filename.end_with? ".rb"
        define_with_json(contents, id)
      end

      def define_with_script(script, id)
        definition = Class.new(A4Tools::ObjectBuilder) {}
        definition.id = id
        definition.class_eval(script, id.to_s)
        definition.new
      end

      def define_with_json(string, id=nil, description=nil)
        object = symbolify(JSON.parse(string))
        description ||= object[:___description] || "Supplied from JSON"
        definition = Class.new(A4Tools::ObjectBuilder) do
          generate { self.class.instance_variable_get("@obj") }
        end
        
        object.delete(:__description)
        definition.class_eval("description \"#{description}\"")
        definition.instance_variable_set("@obj", object)
        definition.id = id.to_sym
        definition.new
      end

      def builders
        @definitions.values
      end

      def builders_for_name(name)
        @definitions.values.select do |defn|
          defn.name_matches?(name)
        end
      end

      def supported_classes(truncate=true)
        r = {}
        @definitions.each do |id, defn|
          key = truncate ? talk.truncated_name(defn.name) : defn.name
          r[key] = true
        end
        r.keys
      end

      def [](identifier)
        @definitions[identifier.to_sym]
      end

      def []=(identifier, defn)
        @definitions[identifier.to_sym] = defn
      end

      def identifiers_like(id)
        @definitions.keys.select { |key| key.to_s.start_with? id.to_s }
      end

      ### DSL stuff for defining items

      def identifier(id)
        @id = id
      end

      def description(desc)
        @description_text = desc
      end

      def cache(is_cached)
        @is_cached = is_cached
      end

      def generate(&block)
        @generate_block = block
      end
    end

    def block_call(&block)
      self.class.send(:define_method, :__block_call, &block)
      self.class.send(:instance_method, :__block_call).bind(self).call
    end

    def identifier
      self.class.id
    end

    def name_matches?(other)
      name.end_with? other.to_s
    end

    def name
      value if @cached.nil?
      @cached[:__class] || ""
    end

    def description
      self.class.description_text
    end

    def value
      return @cached unless @cached.nil? or not self.class.is_cached
      v = block_call(&self.class.generate_block)
      v.delete(:__description)
      @cached = v
    end

    load_objects
  end
end
